#!/usr/bin/env python3

## Stopwords management functions
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Functions :
# - remove_stopwords
# - stopwords_ascii
# - stopwords_nltk
# - stopwords_nltk_ascii


import os
import nltk
import pandas as pd
from typing import Union

from words_n_fun import utils
from words_n_fun.preprocessing import basic

# Get logger
import logging

logger = logging.getLogger(__name__)

# Setting the path where the nltk stopwords data is located
stopwords_dir_path = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))), 'nltk_data')
nltk.data.path.append(stopwords_dir_path)


# From R stopwords package (wrapper around https://github.com/stopwords-iso/stopwords-iso)
STOPWORDS = ["a", "abord", "absolument", "afin", "ah", "ai", "aie", "aient", "aies", "ailleurs", "ainsi", "ait",
             "allaient", "allo", "allons", "allô", "alors", "anterieur", "anterieure", "anterieures", "apres", "après",
             "as", "assez", "attendu", "au", "aucun", "aucune", "aucuns", "aujourd", "aujourd'hui", "aupres", "auquel",
             "aura", "aurai", "auraient", "aurais", "aurait", "auras", "aurez", "auriez", "aurions", "aurons", "auront",
             "aussi", "autre", "autrefois", "autrement", "autres", "autrui", "aux", "auxquelles", "auxquels", "avaient",
             "avais", "avait", "avant", "avec", "avez", "aviez", "avions", "avoir", "avons", "ayant", "ayez", "ayons",
             "b", "bah", "bas", "basee", "bat", "beau", "beaucoup", "bien", "bigre", "bon", "boum", "bravo", "brrr",
             "c", "car", "ce", "ceci", "cela", "celle", "celle-ci", "celle-là", "celles", "celles-ci", "celles-là",
             "celui", "celui-ci", "celui-là", "celà", "cent", "cependant", "certain", "certaine", "certaines",
             "certains", "certes", "ces", "cet", "cette", "ceux", "ceux-ci", "ceux-là", "chacun", "chacune", "chaque",
             "cher", "chers", "chez", "chiche", "chut", "chère", "chères", "ci", "cinq", "cinquantaine", "cinquante",
             "cinquantième", "cinquième", "clac", "clic", "combien", "comme", "comment", "comparable", "comparables",
             "compris", "concernant", "contre", "couic", "crac", "d", "da", "dans", "de", "debout", "dedans", "dehors",
             "deja", "delà", "depuis", "dernier", "derniere", "derriere", "derrière", "des", "desormais", "desquelles",
             "desquels", "dessous", "dessus", "deux", "deuxième", "deuxièmement", "devant", "devers", "devra",
             "devrait", "different", "differentes", "differents", "différent", "différente", "différentes",
             "différents", "dire", "directe", "directement", "dit", "dite", "dits", "divers", "diverse", "diverses",
             "dix", "dix-huit", "dix-neuf", "dix-sept", "dixième", "doit", "doivent", "donc", "dont", "dos", "douze",
             "douzième", "dring", "droite", "du", "duquel", "durant", "dès", "début", "désormais", "e", "effet",
             "egale", "egalement", "egales", "eh", "elle", "elle-même", "elles", "elles-mêmes", "en", "encore", "enfin",
             "entre", "envers", "environ", "es", "essai", "est", "et", "etant", "etc", "etre", "eu", "eue", "eues",
             "euh", "eurent", "eus", "eusse", "eussent", "eusses",  "eussiez", "eussions", "eut", "eux", "eux-mêmes",
             "exactement", "excepté", "extenso", "exterieur", "eûmes", "eût", "eûtes", "f", "fais", "faisaient",
             "faisant", "fait", "faites", "façon", "feront", "fi", "flac", "floc",  "fois", "font", "force", "furent",
             "fus", "fusse", "fussent", "fusses", "fussiez", "fussions", "fut", "fûmes", "fût", "fûtes", "g", "gens",
             "h", "ha", "haut", "hein", "hem", "hep", "hi", "ho", "holà", "hop", "hormis", "hors", "hou", "houp", "hue",
             "hui", "huit", "huitième", "hum", "hurrah", "hé", "hélas", "i", "ici", "il", "ils", "importe", "j", "je",
             "jusqu", "jusque", "juste", "k", "l", "la", "laisser", "laquelle", "las", "le", "lequel", "les",
             "lesquelles", "lesquels", "leur", "leurs", "longtemps", "lors", "lorsque", "lui", "lui-meme", "lui-même",
             "là", "lès", "m", "ma", "maint", "maintenant", "mais", "malgre", "malgré", "maximale", "me", "meme",
             "memes", "merci", "mes", "mien", "mienne", "miennes", "miens", "mille", "mince", "mine", "minimale", "moi",
             "moi-meme", "moi-même", "moindres", "moins", "mon", "mot", "moyennant", "multiple", "multiples", "même",
             "mêmes", "n", "na", "naturel", "naturelle", "naturelles", "ne", "neanmoins", "necessaire",
             "necessairement", "neuf", "neuvième", "ni", "nombreuses", "nombreux", "nommés", "non", "nos", "notamment",
             "notre", "nous", "nous-mêmes", "nouveau", "nouveaux", "nul", "néanmoins", "nôtre", "nôtres", "o", "oh",
             "ohé", "ollé", "olé", "on", "ont", "onze", "onzième", "ore", "ou", "ouf", "ouias", "oust", "ouste",
             "outre", "ouvert", "ouverte", "ouverts", "o", "où", "p", "paf", "pan", "par", "parce", "parfois", "parle",
             "parlent", "parler", "parmi", "parole", "parseme", "partant", "particulier", "particulière",
             "particulièrement", "pas", "passé", "pendant", "pense", "permet", "personne", "personnes", "peu", "peut",
             "peuvent", "peux", "pff", "pfft", "pfut", "pif", "pire", "pièce", "plein", "plouf", "plupart", "plus",
             "plusieurs", "plutôt", "possessif", "possessifs", "possible", "possibles", "pouah", "pour", "pourquoi",
             "pourrais", "pourrait", "pouvait", "prealable", "precisement", "premier", "première", "premièrement",
             "pres", "probable", "probante", "procedant", "proche", "près", "psitt", "pu", "puis", "puisque", "pur",
             "pure", "q", "qu", "quand", "quant", "quant-à-soi", "quanta", "quarante", "quatorze", "quatre",
             "quatre-vingt", "quatrième", "quatrièmement", "que", "quel", "quelconque", "quelle", "quelles",
             "quelqu'un", "quelque", "quelques", "quels", "qui", "quiconque", "quinze", "quoi", "quoique", "r", "rare",
             "rarement", "rares", "relative", "relativement", "remarquable", "rend", "rendre", "restant", "reste",
             "restent", "restrictif", "retour", "revoici", "revoilà", "rien", "s", "sa", "sacrebleu", "sait", "sans",
             "sapristi", "sauf", "se", "sein", "seize", "selon", "semblable", "semblaient", "semble", "semblent",
             "sent", "sept", "septième", "sera", "serai", "seraient", "serais", "serait", "seras", "serez", "seriez",
             "serions", "serons", "seront", "ses", "seul", "seule", "seulement", "si", "sien", "sienne", "siennes",
             "siens", "sinon", "six", "sixième", "soi", "soi-même", "soient", "sois", "soit", "soixante", "sommes",
             "son", "sont", "sous", "souvent", "soyez", "soyons", "specifique", "specifiques", "speculatif", "stop",
             "strictement", "subtiles", "suffisant", "suffisante", "suffit", "suis", "suit", "suivant", "suivante",
             "suivantes", "suivants", "suivre", "sujet", "superpose", "sur", "surtout", "t", "ta", "tac", "tandis",
             "tant", "tardive", "te", "tel", "telle", "tellement", "telles", "tels", "tenant", "tend", "tenir", "tente",
             "tes", "tic", "tien", "tienne", "tiennes", "tiens", "toc", "toi", "toi-même", "ton", "touchant",
             "toujours", "tous", "tout", "toute", "toutefois", "toutes", "treize", "trente", "tres", "trois",
             "troisième", "troisièmement", "trop", "très", "tsoin", "tsouin", "tu", "té", "u", "un", "une", "unes",
             "uniformement", "unique", "uniques", "uns", "v", "va", "vais", "valeur", "vas", "vers", "via", "vif",
             "vifs", "vingt", "vivat", "vive", "vives", "vlan", "voici", "voie", "voient", "voilà", "vont", "vos",
             "votre", "vous", "vous-mêmes", "vu", "vé", "vôtre", "vôtres", "w", "x", "y", "z", "zut", "à", "â", "ça",
             "ès", "étaient", "étais", "était", "étant", "état", "étiez", "étions", "été", "étée", "étées", "étés",
             "êtes", "être", "ô"]

# Specific Pôle Emploi stopwords #1
STOPWORDS_OFFRES_1 = ["recherche", "recherchons", "mission", "missions", "poste", "recrute", "recrutons"]

# Specific Pôle Emploi stopwords #1
STOPWORDS_OFFRES_2 = ["expérience", "assurer", "assurez", "travaux"]


@utils.data_agnostic
@utils.regroup_data_series
def remove_stopwords(docs: pd.Series, opt: str = 'all', set_to_add: Union[list, None] = None,
                     set_to_remove: Union[list, None] = None) -> pd.Series:
    '''Stopwords removal

    Args:
        docs (pd.Series): Documents to process
    Kwargs:
        opt (str): Specifies which stopwords set to use (def='all')
        set_to_add (list): Additionnal stopwords to look for and remove
        set_to_remove (list): Words existing in the stopwords set that should not be removed
    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling stopwords.remove_stopwords')
    if set_to_add is None:
        set_to_add = []
    if set_to_remove is None:
        set_to_remove = []
    # Check if everything is in lowercase (NaNs are replaced, letters are kept)
    if docs.fillna('').str.replace(r"[^A-Za-z]", '').replace('', 'placeholder').str.islower().sum() != docs.shape[0]:
        logger.warning(docs)
        logger.warning('Some characters appear to be in uppercase, stopwords are in lowercase only.')
    # Common soptwords lists
    usage = {
        'none': set(),
        'iso': set().union(STOPWORDS, stopwords_ascii()),
        'nltk': set().union(stopwords_nltk(), stopwords_nltk_ascii()),
        'offres_pe': set().union(STOPWORDS_OFFRES_1, STOPWORDS_OFFRES_2),
        'all': set().union(STOPWORDS, stopwords_ascii(), stopwords_nltk(), stopwords_nltk_ascii()),
    }

    if opt in usage.keys():
        stopwords_list = list(usage.get(opt))
    else:
        logger.warning(
            f"Option {opt} does not exist " +
            f"Existing options are : {', '.join(usage.keys())}. " +
            "By default, all the stopwords are used."
        )
        stopwords_list = STOPWORDS
    # Add custom set
    if len(set_to_add) != 0:
        stopwords_list = list(set(stopwords_list + set_to_add))
    # Remove unwanted words
    if len(set_to_remove) != 0:
        stopwords_list = list(set(stopwords_list) - set(set_to_remove))
    # Empty list case
    if len(stopwords_list) == 0:
        logger.warning("Stopwords_list is empty.")
        logger.warning("Non strings entries are still replaced by None.")
        return docs.apply(lambda x: x if isinstance(x, str) else None)

    regex = utils.get_regex_match_words(stopwords_list)
    return docs.str.replace(regex, '')


def stopwords_ascii() -> list:
    ''' Returns stopwords list in ASCII format (without any special character nor accents)

    Returns:
        list: Stopwords in ASCII format
    '''
    logger.debug('Calling stopwords.stopwords_ascii')
    # Process
    return basic.remove_accents(STOPWORDS, use_tqdm=False)


def stopwords_nltk(try_update: bool = False) -> list:
    ''' Returns the list of FRENCH stopwords from the NLTK package.

    Kwargs:
        try_update (bool): Controls whether we shoud try to update (download required) the stopwords list.
    Returns:
        list: FRENCH stopwords from the NLTK package
    '''
    logger.debug('Calling stopwords.stopwords_nltk')
    if try_update:
        logger.debug("Trying to download an up to date list from NLTK.")
        nltk.download('stopwords', quiet=True)
    return nltk.corpus.stopwords.words('french')


def stopwords_nltk_ascii() -> list:
    ''' Returns the list of FRENCH stopwords from the NLTK package in ASCII format.

    Kwargs:
        try_update (bool): Controls whether we shoud try to update (download required) the stopwords list.
    Returns:
        list: FRENCH stopwords from the NLTK package in ASCII format
    '''
    logger.debug('Calling stopwords.stopwords_nltk_ascii')
    # Process
    return basic.remove_accents(stopwords_nltk(), use_tqdm=False)


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3

## Functions linked to the lemmatizer
#  It used to be a custom made french lemmatizer but has been depcrecated and replaced by spacy
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Fonctions :
# - lemmatize -> Lemmatizes text


import sys
import requests
import pandas as pd
import simplejson as json
from words_n_fun import utils

# Get logger
import logging
logger = logging.getLogger(__name__)


# Spacy has to be installed for the lemmatizer to work
# Since it is an optional dependency, a warning is raised if a call to the lemmatizer is done
if 'spacy' in sys.modules:
    LEMMATIZER_AVAILABLE = True
    import spacy
    try:
        if not spacy.util.is_package("fr_core_news_sm"):
            logger.info("Downloading fr_core_news_sm")
            spacy.cli.download('fr_core_news_sm')
        spacy_model = spacy.load('fr_core_news_sm')
    except:
        spacy_model = None
        logger.error("Unexpected error downloading or loading spacy model fr_core_news_sm")
else:
    LEMMATIZER_AVAILABLE = False
    logger.warning("Spacy has not been found, lemmatizer features are not available.")
    logger.warning("To use it, you must install spacy. For instance: pip install words-n-fun[lemmatizer]")

@utils.data_agnostic
@utils.regroup_data_series
def lemmatize(docs: pd.Series) -> pd.Series:
    '''Text lemmatizer - spacy
    #   This feature uses the fr_core_news_sm from Spacy to process the text

    Args:
        docs (pd.Series): Documents to process

    Raises:
        ImportError : If spacy is not found
        Exception : fr_core_news_sm model is not found
    Returns:
        pd.Series: Modified documents
    '''
    if not LEMMATIZER_AVAILABLE:
        logger.error("Spacy has not been found, lemmatizer features are not available.")
        logger.error("To use it, you must install spacy. For instance: pip install words-n-fun[lemmatizer]")
        raise ImportError("Spacy has not been found, lemmatizer features are not available.")
    if not spacy.util.is_package("fr_core_news_sm"):
        logger.error("Unable to call spacy lemmatizer withouth spacy fr_core_news_sm model")
        raise Exception("Unable to call spacy lemmatizer withouth spacy fr_core_news_sm model")
    docs = (
        pd.Series(docs)
        .str.lower()
        .str.replace('\W', ' ')
        .str.replace(r"([0-9]+(\.[0-9]+)?)", r" \1 ")
        .str.replace('\s+', ' ', regex=True)
        .str.strip()
    )
    docs = list(docs.values)
    if len(docs) == 0:
        return None
    lemmatized = []
    for doc in docs:
        if isinstance(doc, str):
            content = spacy_model(doc)
            lem = ' '.join([token.lemma_ for token in content])
            lemmatized.append(lem)
        else:
            lemmatized.append(None)

    return pd.Series(lemmatized)



if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3

## Main funtions of the preprocessing API
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Fonctions :
# - notnull -> Replaces null values by an empty character
# - remove_non_string -> Replaces all non strings by an empty character
# - get_true_spaces -> Replaces all whitespaces by a single space
# - to_lower -> Transforms the string to lower case
# - pe_matching -> Specific one-to-one tokens replacements
# - remove_punct -> Replaces all non alpha-numeric characters by spaces
# - trim_string -> Trims spaces at the beginning and ending of the string (multiple spaces become one)
# - remove_numeric -> Replaces numeric strings by a space
# - remove_stopwords -> Removes stopwords
# - remove_accents -> Removes all accents and special characters (ç..)
# - remove_gender_synonyms -> [French] Removes gendered synonyms
# - lemmatize -> Lemmatizes the document
# - stemmatize -> Stemmatizes the words of the document
# - add_point -> Adds a dot at the end of each line
# - deal_with_specific_characters -> Ads spaces before and after some punctuations (, : ; .)
# - replace_urls -> Replaces URLs by spaces
# - remove_words -> Replaces words from a list
# - fix_text -> Fixes numerous inconsistencies within a text (via ftfy)


import ftfy
import unicodedata
import pandas as pd
from typing import Union, List
from nltk.stem.snowball import FrenchStemmer

from words_n_fun import utils
from words_n_fun import CustomTqdm as tqdm
from words_n_fun.preprocessing import stopwords
from words_n_fun.preprocessing import lemmatizer
from words_n_fun.preprocessing import synonym_malefemale_replacement


tqdm.pandas()

# Get logger
import logging

logger = logging.getLogger(__name__)


@utils.data_agnostic
def notnull(docs: pd.Series) -> pd.Series:
    '''Replaces null values by an empty character

    Args:
        docs (pd.Series): Documents to process

    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.notnull')
    return docs.fillna('')


@utils.data_agnostic
def remove_non_string(docs: pd.Series) -> pd.Series:
    '''Replaces all non strings by an empty character

    Args:
        docs (pd.Series): Documents to process

    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.remove_non_string')
    return docs.progress_apply(lambda x: x if isinstance(x, str) else '')


@utils.data_agnostic
@utils.regroup_data_series
def get_true_spaces(docs: pd.Series) -> pd.Series:
    '''Replaces all whitespaces by a single space

    Args:
        docs (pd.Series): Documents to process

    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.get_true_spaces')
    return docs.str.replace(r'\s', ' ')


@utils.data_agnostic
@utils.regroup_data_series
def to_lower(docs: pd.Series, threshold_nb_chars: int = 0) -> pd.Series:
    '''Transforms the string to lower case

    Args:
        docs (pd.Series): Documents to process
    Kwargs:
        threshold_nb_chars (int): Minimum number of characters for a token to be transformed to lowercase (def=0).

    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.to_lower')
    if threshold_nb_chars > 1:
        logger.debug(f"Applying lower case transform for tokens of at least {threshold_nb_chars} chars.")
        return docs.progress_apply(lambda x: " ".join(x.lower() if len(x) >= threshold_nb_chars else x for x in x.split((" "))) if isinstance(x, str) else None)
    else:
        return docs.str.lower()


@utils.data_agnostic
@utils.regroup_data_series
def pe_matching(docs: pd.Series) -> pd.Series:
    '''Specific one-to-one tokens replacements
    For instance 'Permis b' => 'Permisb'

    Args:
        docs (pd.Series): Documents to process

    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.pe_matching')
    # One can add more rules here
    regex = utils.get_regex_match_words(['(permis)\s+(b)'], case_insensitive=True, words_as_regex=True)
    docs = docs.str.replace(regex, r'\2\3')
    return docs


@utils.data_agnostic
@utils.regroup_data_series
def remove_punct(docs: pd.Series, del_parenthesis: bool = True, replacement_char: str = ' ') -> pd.Series:
    '''Replaces all non alpha-numeric characters by spaces

    Args:
        docs (pd.Series): Documents to process
    Kwargs:
        del_parenthesis (bool): Whether parenthesis and slashes are removed (def= True)
        replacement_char (str): Replacement character (def= ' ')
    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.remove_punct')
    if not del_parenthesis:
        regex = r"[^\w\s\(\)\/]|_"
    else:
        regex = r"[^\w\s]|_"
    return docs.str.replace(regex, replacement_char)


@utils.data_agnostic
@utils.regroup_data_series
def trim_string(docs: pd.Series) -> pd.Series:
    '''Trims spaces: multiple spaces become one

    Args:
        docs (pd.Series): Documents to process

    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.trim_string')
    # TODO: better way ?
    docs = docs.str.replace(r'[\t\f\v ]{2,}', ' ')
    docs = remove_leading_and_ending_spaces(docs)
    return docs


@utils.data_agnostic
@utils.regroup_data_series
def remove_leading_and_ending_spaces(docs: pd.Series) -> pd.Series:
    '''Removes leading and trailing spaces

    Args:
        docs (pd.Series): Documents to process

    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.remove_leading_and_ending_spaces')
    docs = docs.str.replace(r'^(\s)+', '')
    return docs.str.replace(r'(\s)+$', '')


@utils.data_agnostic
@utils.regroup_data_series
def remove_numeric(docs: pd.Series, replacement_char: str = ' ') -> pd.Series:
    '''Replaces numeric strings by a space

    Args:
        docs (pd.Series): Documents to process
    Kwargs:
        replacement_char (str): Replacement character (def= ' ')
    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.remove_numeric')
    return docs.str.replace(r'([0-9]+)', replacement_char)


@utils.data_agnostic
@utils.regroup_data_series
def remove_stopwords(docs: pd.Series, opt: str = 'all', set_to_add: Union[list, None] = None,
                     set_to_remove: Union[list, None] = None) -> pd.Series:
    '''Removes stopwords

    Args:
        docs (pd.Series): Documents to process
    Kwargs:
        opt (str): Specifies which stopwords set are used, cf stopwords.py (def='all')
        set_to_add (list): List of words to append to the stopwords list
        set_to_remove (list): List of words to remove from the stopwords list
    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.remove_stopwords')
    if set_to_add is None:
        set_to_add = []
    if set_to_remove is None:
        set_to_remove = []
    return stopwords.remove_stopwords(docs, opt=opt, set_to_add=set_to_add, set_to_remove=set_to_remove)


@utils.data_agnostic
@utils.regroup_data_series
def remove_accents(docs: pd.Series, use_tqdm: bool = True) -> pd.Series:
    '''Removes all accents and special characters (ç..)

    Args:
        docs (pd.Series): Documents to process
    Kwargs:
        use_tqdm (bool): Whether tqdm should be used (default: True)

    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.remove_accents')
    if use_tqdm:
        return docs.progress_apply(lambda x: ''.join((c for c in unicodedata.normalize('NFD', x) if unicodedata.category(c) != 'Mn')) if isinstance(x, str) else None)
    else:
        return docs.apply(lambda x: ''.join((c for c in unicodedata.normalize('NFD', x) if unicodedata.category(c) != 'Mn')) if isinstance(x, str) else None)


@utils.data_agnostic
@utils.regroup_data_series
def remove_gender_synonyms(docs: pd.Series) -> pd.Series:
    '''[French] Removes gendered synonyms
    # Find occurences such as "male version / female version" (eg: Coiffeur / Coiffeuse)
    # By convention, the male version is kept (in accordance with the lemmatizer)

    Args:
        docs (pd.Series): Documents to process

    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.remove_gender_synonyms')
    return synonym_malefemale_replacement.remove_gender_synonyms(docs)


@utils.data_agnostic
@utils.regroup_data_series
def lemmatize(docs: pd.Series) -> pd.Series:
    '''Lemmatizes the documents
    Appel à une API externe

    Args:
        docs (pd.Series): Documents to process

    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.lemmatize')
    # Process
    return lemmatizer.lemmatize(docs)


@utils.data_agnostic
@utils.regroup_data_series
def stemmatize(docs: pd.Series) -> pd.Series:
    '''Stemmatizes words in the documents

    Args:
        docs (pd.Series): Documents to process

    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.stemmatize')
    logger.warning('Calling the FRENCH stemmer')
    stemmer = FrenchStemmer()
    return docs.progress_apply(lambda x: " ".join(stemmer.stem(x) for x in x.split(' ')) if isinstance(x, str) else None)


@utils.data_agnostic
@utils.regroup_data_series
def add_point(docs: pd.Series) -> pd.Series:
    '''Adds a dot at the end of each line

    Args:
        docs (pd.Series): Documents to process

    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.add_point')
    return docs.progress_apply(lambda x: (x + '.' if not x.endswith('.') else x) if isinstance(x, str) else None)


@utils.data_agnostic
@utils.regroup_data_series
def deal_with_specific_characters(docs: pd.Series) -> pd.Series:
    '''Adds spaces before and after some punctuations (, : ; .)

    Args:
      docs (pd.Series): Documents to process
    Returns:
      pd.Series: Modified documents
    '''
    logger.debug('Calling basic.deal_with_specific_characters')
    return docs.str.replace(r"(\s)?([',.;:])(\s)?", r' \2 ')


@utils.data_agnostic
@utils.regroup_data_series
def replace_urls(docs: pd.Series, replacement_char: str = ' ', replace_with_domain: bool = False) -> pd.Series:
    '''Replaces URLs by either a str or the url domain

    Args:
        docs (pd.Series): Documents to process
    Kwargs:
        replacement_char (str): Replacement character (def= ' ')
        replace_with_domain (bool): Replacement_char is overriden and the url is replaced by its domain (def= False)
    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.replace_urls')
    # based on : https://stackoverflow.com/questions/6038061/regular-expression-to-find-urls-within-a-string
    regex = r'(?i)(?<!\w|/)(((http|ftp|https):\/\/)*(www\.|ftp\.)+|((http|ftp|https):\/\/)+(www\.|ftp\.)*)([\w_-]+(?:(?:\.[\w_-]+)+))([\w.,@?^=%&:\/~+#-]*[\w@?^=%&\/~+#-])?'
    if not replace_with_domain:
        return docs.str.replace(regex, replacement_char)
    else:
        return docs.str.replace(regex, r' \8 ')


@utils.data_agnostic
@utils.regroup_data_series
def remove_words(docs: pd.Series, words_to_remove: List[str], case_insensitive=False) -> pd.Series:
    '''Function to remove words from a list

    Args:
        docs (pd.Series): Documents to process
        words_to_remove (list<str>): List of words to remove
    Kwargs:
        case_insensitive (bool): Whether the replacement is case sensitive (defaut : False)
    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling utils.remove_words')
    regex = utils.get_regex_match_words(words_to_remove, case_insensitive=case_insensitive)
    return docs.str.replace(regex, '')


@utils.data_agnostic
@utils.regroup_data_series
def fix_text(docs: pd.Series, **ftfy_kwargs) -> pd.Series:
    '''Fixes numerous inconsistencies within a text (via ftfy)
       By default:
        - Removes some Linux instructions
        - Fixes encoding
        - Fixes HTML entities
        - Fixes some quotes
        - Replaces tied letter (e.g. œ)
        - Replaces characters larger than normal
        - Fixes line breaks (LF)
        - Fixes UTF-16 "surrogates"
        - Removes "control characters"
        - Removes "Byte-Order Mark"
        - NFC Normalization (accents)

    Args:
        docs (pd.Series): Documents to process
    Kwargs:
        ftfy_kwargs (dict): Kwargs forwarded to ftfy

    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling basic.fix_text')
    return docs.progress_apply(lambda x: ftfy.fix_text(x, **ftfy_kwargs) if isinstance(x, str) else None)


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3

## Functions to remove gendered synonyms: incredibly specific to the french language
# TODO : To remove?
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Fonctions :
# - remove_gender_synonyms -> Removes gendered synonyms
# - matching_words -> Male/Female token matching
# - update_synonyms_set -> Update the synonyms set


import re
import pandas as pd
from typing import Tuple

from words_n_fun import utils

# Get logger
import logging

logger = logging.getLogger(__name__)


SYNONYM_DICT = {
    'eur': ['euse', 'se', 'rice', 'e', 'eure'],
    "ien": ['ne', 'nne', 'e'],
    "ier": ['ere', 're', 'e'],
    "ial": ['e'],
    "l": ['le'],
    "nt": ['e'],
    "at": ['e'],
    "f": ['ve'],
    "o": ['a'],
    "er": ['ere', 'ère', 'euse'],
    "i": ['e'],
    "d": ['e'],
    "é": ['e'],
    "e": ['e'],
    "on": ['onne'],
    "is": ['ise'],
    "s": ['s']
}


@utils.data_agnostic
@utils.regroup_data_series
def remove_gender_synonyms(docs: pd.Series) -> pd.Series:
    '''Removes gendered synonyms

    Args:
        docs (pd.Series): Documents to process
    Returns:
        pd.Series: Modified documents
    '''
    logger.debug('Calling synonym_malefemale_replacement.getSynonyms')

    ## Preprocessing
    docs = docs.str.replace('(\s*)/(\s*)', '/')  # Removes whitespaces around "/"
    docs = docs.str.replace('(\s*)\((\s*)', '(')  # Removes potential whitespaces before "("
    docs = docs.str.replace('\)', ') ')  # Add a space after ")"

    ## Set match paterns
    parenthesis_pattern = r"([\w\-]+)\(([\w\-]+)\)()"  # Case :  serveur(se)
    slash_pattern = r"([\w\-]+)/([\w\-]+)(\([\w\-]+\))?"  # Case: serveur/serveuse and serveur/serveur(se)
    slash_pattern_BiWords = r"([\w\-]+\s[\w\-]+)/([\w\-]+\s[\w\-]+)()"  # Case: apprenti boucher/apprentie bouchere
    slash_pattern_TriWords = r"([\w\-]+\s[\w\-]+\s[\w\-]+)/([\w\-]+\s[\w\-]+\s[\w\-]+)()"  # Case:  aide apprenti boucher/aide apprentie bouchere
    match_pattern = [parenthesis_pattern, slash_pattern, slash_pattern_BiWords, slash_pattern_TriWords]

    ## Creating synonyms listings
    synonyms_set = {}
    for i in range(len(docs)):  # For every document
        for match in match_pattern:  # ... and for every match ...
            # Update the synonyms set
            text = docs.iloc[i]
            if isinstance(text, str):
                synonyms_set = update_synonyms_set(synonyms_set, re.findall(match, text), i)

    ## Update the  documents
    for i in range(len(docs)):  # For every document
        text = docs.iloc[i]

        if not isinstance(text, str):
            text = None
        else:
            # We process each pattern individually
            match_parenthesis_pattern = re.findall(parenthesis_pattern, text)
            match_slash_pattern = re.findall(slash_pattern, text)
            match_slash_pattern_BiWords = re.findall(slash_pattern_BiWords, text)
            match_slash_pattern_TriWords = re.findall(slash_pattern_TriWords, text)

            # Parenthesis
            if len(match_parenthesis_pattern) != 0:
                for (word1, word2, word3) in match_parenthesis_pattern:
                    if (word1, word2, word1) in synonyms_set:
                        text = text.replace(word1 + "(" + word2 + ")", word1)  # Case: serveur(se)
            # Slashes
            if len(match_slash_pattern) != 0:
                for (word1, word2, word3) in match_slash_pattern:
                    if word1 == word2:  # Case: serveur/serveur(se)
                        text = text.replace(word1 + "/" + word2 + "(" + word3 + ")", word1)
                        text = text.replace(word1 + "/" + word2, word1)
                    elif (word1, word2, word1) in synonyms_set:
                        text = text.replace(word1 + "/" + word2 + "(" + word3 + ")", word1)
                        text = text.replace(word1 + "/" + word2, word1)
            # Slashes BiWords
            if len(match_slash_pattern_BiWords) != 0:
                for (word1, word2, word3) in match_slash_pattern_BiWords:  # Case: apprenti boucher/apprentie bouchere
                    if (word1, word2, word1) in synonyms_set:
                        text = text.replace(word1 + "/" + word2, word1)
            # Slashes TriWords
            if len(match_slash_pattern_TriWords) != 0:
                for (word1, word2, word3) in match_slash_pattern_TriWords:
                    if (word1, word2, word1) in synonyms_set:
                        text = text.replace(word1 + "/" + word2, word1)
            docs.iloc[i] = text
    return docs


def matching_words(word1: str, word2: str) -> Tuple[str, str, str]:
    '''Male/Female token matching

    Args:
        word1 (str)
        word2 (str)

    Raises:
        TypeError : If word1 is empty
        TypeError : If word2 is empty

    Returns:
        tuple: (word1,word2,word1) if there is a match
        tuple: (word1,word2,"unknown") otherwise

    Rules :
        - If the end of word1 exists in the dictionnary and the end of word2 matches one of its values: eg serveur/se
        - If the end of word1 exists in the dictionnary and the end of word2 matches one of its values and the 4 first letters of both word1 and word 2 are identical: eg serveur/serveuse
        - If the end of word1 exists in the dictionnary and the end of word2 matches one of its values and the length of word2 is lesser or equal than the length of word1: eg conducteur(trice)
    '''

    if len(word1) == 0:
        raise TypeError('Word1 is empty.')
    if len(word2) == 0:
        raise TypeError('Word2 is empty.')

    combi = None
    for word1_end in SYNONYM_DICT.keys():
        for word2_end in SYNONYM_DICT[word1_end]:
            if combi is None:
                if word1[-len(word1_end) :] == word1_end:
                    if (
                        (word2 == word2_end)
                        or (word2[-len(word2_end) :] == word2_end and word1[:4] == word2[:4])
                        or (word2[-len(word2_end) :] == word2_end and len(word2) <= (len(word2_end) + 2))
                    ):
                        combi = (word1, word2, word1)
    if combi is None:
        combi = (word1, word2, "unknown")
    return combi


def update_synonyms_set(synonyms_set: dict, match: list, numligne: int) -> dict:
    '''Update the synonyms set

    Args:
        synonyms_set (dict)
        match (liste of tuples), eg : [('serveur', 'serveur', '(se)'), ('boucher', 'boucher', '(ere)')]
        numligne (int)
    Returns:
        dict : Synonyms set
    '''
    if len(match) == 0:
        pass
    else:
        combi = None
        for (word1, word2, word3) in match:
            # Example : Case serveur/serveur(se) #re.match("[^\d\s]+$", word1) : word1 is neither digit nor whitespace
            if word1 == word2 and re.match("^[\d\s]+$", word1) is None:
                combi = (word1, word2, word1)
            else:
                combi = matching_words(word1, word2)
            if combi is not None:
                if (list(combi)[2]) != "unknown" and combi not in synonyms_set.keys():
                    synonyms_set[combi] = numligne
    return synonyms_set


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3

## Functions to split texts into sentences
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# Fonctions :
# - split_sentences : Splits a text into sentences
# - split_sentences_df : Splits a set of texts into sentences


import re
import pandas as pd
from itertools import chain
from typing import Union, List

from words_n_fun import CustomTqdm as tqdm


tqdm.pandas()

# Get logger
import logging

logger = logging.getLogger(__name__)


def split_sentences(text: str) -> List[str]:
    '''Splits a text into sentences

    Args:
        text (str): Arbitrary text
    Returns:
        list<str> : List of sentences
    '''

    # We split around \n after singling them out and removing trailing spaces.
    text = re.compile("\s*\n+").sub('\n', text)
    text = re.compile("\s+$").sub('', text)
    text_list = text.split('\n')

    # Text between parenthesis are kept aside
    parentheses_matches = {}
    parentheses_regex = r"\(([^)]+)\)"
    k = 0  # Incrementation UID
    for i, sentence in enumerate(text_list):
        for match in re.compile(parentheses_regex).finditer(sentence):
            tmp_text = match.group()
            tmp_replace = f"UID_PE_PARENTHESES_{k}"
            parentheses_matches[tmp_replace] = tmp_text
            sentence = sentence.replace(tmp_text, tmp_replace, 1)
            text_list[i] = sentence
            k += 1

    # We split around ".", "!", "?" if they are followed by a whitespace or a newline
    # Special occurences ("i.e", "e.g", "M.") are skipped
    regex = "(?<!\s\w\.\w.\s)(?<!\sM\.\s)(?<=(?:\.|\?|\!)\s)"
    text_list = list(chain.from_iterable([re.compile(regex).split(_) for _ in text_list]))

    # Parenthesis texts are injected back in the list
    for i, elem in enumerate(text_list):
        for k in parentheses_matches.keys():
            elem = elem.replace(k, parentheses_matches[k])
        text_list[i] = elem

    return text_list


def split_sentences_df(df: pd.DataFrame, col: Union[str, int]) -> pd.DataFrame:
    '''Function to split several texts from a pandas DataFrame into sentences

    Args:
        df (pd.DataFrame): DataFrame containing the texts
        col (str ou int): Column name where the text is
    Returns:
        pd.DataFrame: New DataFrame with the text split into sentences
    '''
    df2 = df.copy()
    df2[col] = df2[col].progress_apply(split_sentences)
    return df2.explode(col).reset_index(drop=True)


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")
